/*
 * Copyright 2014 Canonical Ltd.
 *
 * This file is part of webbrowser-app.
 *
 * webbrowser-app is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * webbrowser-app is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

// ==UserScript==
// @run-at document-start
// ==/UserScript==

// Ensure that so-called "smart banners" generated by the smartbanner jQuery
// plugin (http://jasny.github.io/jquery.smartbanner/) that advertise native
// Android/iOS applications based on a very naïve parsing of the UA string
// are never shown.

var exdate = new Date();
exdate.setFullYear(exdate.getFullYear() + 5);
try {
    document.cookie = "sb-closed=true;path=/;expires=" + exdate.toUTCString();
} catch (e) {}
